@@ -14,7 +14,7 @@
  *  limitations under the License.
  *
  */
-#include "luv.h"
+#include "private.h"
 
 static uv_udp_t* luv_check_udp(lua_State* L, int index) {
   uv_udp_t* handle = (uv_udp_t*)luv_checkudata(L, index, "uv_udp");
@@ -27,31 +27,87 @@ static int luv_new_udp(lua_State* L) {
   lua_settop(L, 1);
   uv_udp_t* handle = (uv_udp_t*)luv_newuserdata(L, sizeof(*handle));
   int ret;
-  if (lua_isnoneornil(L, 1)) {
-    ret = uv_udp_init(ctx->loop, handle);
-  }
-  else {
-    unsigned int flags = AF_UNSPEC;
+#if LUV_UV_VERSION_GEQ(1, 39, 0)
+  // TODO: This default can potentially be increased, but it's
+  //       not clear what the best default would be, or if unconditionally
+  //       using recvmmsg is always an improvement.
+  //
+  //       Would probably need to do some extensive benchmarking to
+  //       figure out what a good default might be.
+  int mmsg_num_msgs = 1;
+#endif
+#if LUV_UV_VERSION_GEQ(1, 7, 0)
+  unsigned int flags = AF_UNSPEC;
+  if (!lua_isnoneornil(L, 1)) {
     if (lua_isnumber(L, 1)) {
       flags = lua_tointeger(L, 1);
     }
     else if (lua_isstring(L, 1)) {
       const char* family = lua_tostring(L, 1);
       flags = luv_af_string_to_num(family);
-      if (!flags) {
-        luaL_argerror(L, 1, lua_pushfstring(L, "invalid or unknown address family: '%s'", family));
+    } else if (lua_istable(L, 1)) {
+      lua_getfield(L, 1, "family");
+      if (lua_isnumber(L, -1)) {
+        // The lower 8 bits of the flags parameter are used as the socket domain
+        flags = lua_tointeger(L, -1) & 0xFF;
       }
+      else if (lua_isstring(L, -1)) {
+        flags = luv_af_string_to_num(lua_tostring(L, -1));
+      }
+      else if (!lua_isnil(L, -1)) {
+        luaL_argerror(L, 1, "family must be string or integer if set");
+      }
+      lua_pop(L, 1);
+
+#if LUV_UV_VERSION_GEQ(1, 39, 0)
+      lua_getfield(L, 1, "mmsgs");
+      if (lua_isnumber(L, -1)) {
+        mmsg_num_msgs = lua_tonumber(L, -1);
+      } else if (!lua_isnil(L, -1)) {
+        luaL_argerror(L, 1, "mmsgs must be integer if set");
+      }
+      lua_pop(L, 1);
+#endif
     }
     else {
-      luaL_argerror(L, 1, "expected string or integer");
+      luaL_argerror(L, 1, "expected table, string, or integer");
     }
-    ret = uv_udp_init_ex(ctx->loop, handle, flags);
   }
+#if LUV_UV_VERSION_GEQ(1, 39, 0)
+  // Libuv intended to enable this by default, but it caused a backwards-incompatibility with how
+  // the buffer is freed in udp_recv_cb, so it had to be put behind a flag to avoid breaking
+  // existing libuv users. However, because luv handles UV_UDP_MMSG_CHUNK in luv_udp_recv_cb, we can
+  // always enable this flag and get the benefits of recvmmsg for platforms that support it.
+  //
+  // Relevant links:
+  // - https://github.com/libuv/libuv/issues/2791
+  // - https://github.com/libuv/libuv/pull/2792
+  // - https://github.com/libuv/libuv/pull/2532
+  // - https://github.com/libuv/libuv/issues/419
+  //
+  // But we should only set the flag if we can actually take advantage of it.
+  if (mmsg_num_msgs > 1)
+    flags |= UV_UDP_RECVMMSG;
+#endif
+  ret = uv_udp_init_ex(ctx->loop, handle, flags);
+#else
+  ret = uv_udp_init(ctx->loop, handle);
+#endif
   if (ret < 0) {
     lua_pop(L, 1);
     return luv_error(L, ret);
   }
   handle->data = luv_setup_handle(L, ctx);
+#if LUV_UV_VERSION_GEQ(1, 39, 0)
+  if (flags & UV_UDP_RECVMMSG) {
+    // store the number of msgs to be received for use in alloc_cb
+    int* extra_data = malloc(sizeof(int));
+    assert(extra_data);
+    *extra_data = mmsg_num_msgs;
+    ((luv_handle_t*)handle->data)->extra = extra_data;
+    ((luv_handle_t*)handle->data)->extra_gc = free;
+  }
+#endif
   return 1;
 }
 
@@ -116,12 +172,26 @@ static const char *const luv_membership_opts[] = {
 static int luv_udp_set_membership(lua_State* L) {
   uv_udp_t* handle = luv_check_udp(L, 1);
   const char* multicast_addr = luaL_checkstring(L, 2);
-  const char* interface_addr = luaL_checkstring(L, 3);
+  const char* interface_addr = lua_isstring(L, 3) ? lua_tostring(L, 3) : NULL;
+  luaL_argcheck(L, lua_isstring(L, 3) || lua_isnil(L, 3), 3, "expected string or nil");
   uv_membership membership = (uv_membership)luaL_checkoption(L, 4, NULL, luv_membership_opts);
   int ret = uv_udp_set_membership(handle, multicast_addr, interface_addr, membership);
   return luv_result(L, ret);
 }
 
+#if LUV_UV_VERSION_GEQ(1, 32, 0)
+static int luv_udp_set_source_membership(lua_State* L) {
+  uv_udp_t* handle = luv_check_udp(L, 1);
+  const char* multicast_addr = luaL_checkstring(L, 2);
+  const char* interface_addr = lua_isstring(L, 3) ? lua_tostring(L, 3) : NULL;
+  luaL_argcheck(L, lua_isstring(L, 3) || lua_isnil(L, 3), 3, "expected string or nil");
+  const char* source_addr = luaL_checkstring(L, 4);
+  uv_membership membership = (uv_membership)luaL_checkoption(L, 5, NULL, luv_membership_opts);
+  int ret = uv_udp_set_source_membership(handle, multicast_addr, interface_addr, source_addr, membership);
+  return luv_result(L, ret);
+}
+#endif
+
 static int luv_udp_set_multicast_loop(lua_State* L) {
   uv_udp_t* handle = luv_check_udp(L, 1);
   int on, ret;
@@ -218,42 +288,37 @@ static struct sockaddr* luv_check_addr(lua_State *L, struct sockaddr_storage* ad
 static int luv_udp_send(lua_State* L) {
   uv_udp_t* handle = luv_check_udp(L, 1);
   uv_udp_send_t* req;
-  uv_buf_t buf;
   int ret, ref;
   struct sockaddr_storage addr;
   struct sockaddr* addr_ptr;
   luv_handle_t* lhandle = handle->data;
-
-  luv_check_buf(L, 2, &buf);
   addr_ptr = luv_check_addr(L, &addr, 3, 4);
   ref = luv_check_continuation(L, 5);
   req = (uv_udp_send_t*)lua_newuserdata(L, sizeof(*req));
   req->data = luv_setup_req(L, lhandle->ctx, ref);
-
-  ret = uv_udp_send(req, handle, &buf, 1, addr_ptr, luv_udp_send_cb);
+  size_t count;
+  uv_buf_t* bufs = luv_check_bufs(L, 2, &count, (luv_req_t*)req->data);
+  ret = uv_udp_send(req, handle, bufs, count, addr_ptr, luv_udp_send_cb);
+  free(bufs);
   if (ret < 0) {
     luv_cleanup_req(L, (luv_req_t*)req->data);
     lua_pop(L, 1);
     return luv_error(L, ret);
   }
-
-  lua_pushvalue(L, 2);
-  ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-  lua_pop(L, 1);
-
   lua_pushinteger(L, ret);
   return 1;
 }
 
 static int luv_udp_try_send(lua_State* L) {
   uv_udp_t* handle = luv_check_udp(L, 1);
-  uv_buf_t buf;
   int err_or_num_bytes;
   struct sockaddr_storage addr;
   struct sockaddr* addr_ptr;
-  luv_check_buf(L, 2, &buf);
+  size_t count;
+  uv_buf_t* bufs = luv_check_bufs_noref(L, 2, &count);
   addr_ptr = luv_check_addr(L, &addr, 3, 4);
-  err_or_num_bytes = uv_udp_try_send(handle, &buf, 1, addr_ptr);
+  err_or_num_bytes = uv_udp_try_send(handle, bufs, count, addr_ptr);
+  free(bufs);
   if (err_or_num_bytes < 0) return luv_error(L, err_or_num_bytes);
   lua_pushinteger(L, err_or_num_bytes);
   return 1;
@@ -263,6 +328,16 @@ static void luv_udp_recv_cb(uv_udp_t* handle, ssize_t nread, const uv_buf_t* buf
   luv_handle_t* data = (luv_handle_t*)handle->data;
   lua_State* L = data->ctx->L;
 
+#if LUV_UV_VERSION_GEQ(1, 40, 0)
+  // If UV_UDP_MMSG_FREE is set, we can skip calling the callback
+  // and return early because we know the only purpose of this recv_cb call
+  // is to free the buffer that was being used by recvmmsg
+  if (flags & UV_UDP_MMSG_FREE) {
+    free(buf->base);
+    return;
+  }
+#endif
+
   // err
   if (nread < 0) {
     luv_status(L, nread);
@@ -286,7 +361,15 @@ static void luv_udp_recv_cb(uv_udp_t* handle, ssize_t nread, const uv_buf_t* buf
   else {
     lua_pushnil(L);
   }
+#if LUV_UV_VERSION_GEQ(1, 35, 0)
+  // UV_UDP_MMSG_CHUNK Indicates that the message was received by recvmmsg, so the buffer provided
+  // must not be freed by the recv_cb callback.
+  if (buf && !(flags & UV_UDP_MMSG_CHUNK)) {
+    free(buf->base);
+  }
+#else
   if (buf) free(buf->base);
+#endif
 
   // address
   if (addr) {
@@ -302,18 +385,46 @@ static void luv_udp_recv_cb(uv_udp_t* handle, ssize_t nread, const uv_buf_t* buf
     lua_pushboolean(L, 1);
     lua_setfield(L, -2, "partial");
   }
+#if LUV_UV_VERSION_GEQ(1, 35, 0)
+  if (flags & UV_UDP_MMSG_CHUNK) {
+    lua_pushboolean(L, 1);
+    lua_setfield(L, -2, "mmsg_chunk");
+  }
+#endif
 
   luv_call_callback(L, (luv_handle_t*)handle->data, LUV_RECV, 4);
 }
 
+#if LUV_UV_VERSION_GEQ(1, 39, 0)
+#define MAX_DGRAM_SIZE (64*1024)
+
+static void luv_udp_alloc_cb(uv_handle_t* handle, size_t suggested_size, uv_buf_t* buf) {
+  size_t buffer_size = suggested_size;
+  if (uv_udp_using_recvmmsg((uv_udp_t*)handle)) {
+    int num_msgs = *(int*)(((luv_handle_t*)handle->data)->extra);
+    buffer_size = MAX_DGRAM_SIZE * num_msgs;
+  }
+  buf->base = (char*)malloc(buffer_size);
+  assert(buf->base);
+  buf->len = buffer_size;
+}
+#endif
+
 static int luv_udp_recv_start(lua_State* L) {
   uv_udp_t* handle = luv_check_udp(L, 1);
   int ret;
   luv_check_callback(L, (luv_handle_t*)handle->data, LUV_RECV, 2);
+#if LUV_UV_VERSION_GEQ(1, 39, 0)
+  ret = uv_udp_recv_start(handle, luv_udp_alloc_cb, luv_udp_recv_cb);
+#else
   ret = uv_udp_recv_start(handle, luv_alloc_cb, luv_udp_recv_cb);
+#endif
 #if LUV_UV_VERSION_LEQ(1, 23, 0)
+#if LUV_UV_VERSION_GEQ(1, 10, 0)
   // in Libuv <= 1.23.0, uv_udp_recv_start will return untranslated error codes on Windows
+  // (uv_translate_sys_error was only made public in libuv >= 1.10.0)
   ret = uv_translate_sys_error(ret);
+#endif
 #endif
   return luv_result(L, ret);
 }
